package Utility;

use Exporter;
@ISA = qw(Exporter);

@EXPORT = qw(
  absoluteFileName absoluteReadLink alias allBinaries allFilesReady
  appendBeforeGZ arch await estimatedFinishTime fileToHash fileToList
  filesFromSmallest hashToFile identicalFiles isComplete listToFile max
  min mkdir_p moveAside newer nonZeroBinaries openFromPaste openFromPipe
  openToAppend openToPipe openToRead openToWrite readAliases
  replaceBeforeGZ report sameList stamp touch unstamp readFastaFile
  writeFastaFile avoidInitialDigit printGFF readQuad
);

use FileHandle;
use Sys::Hostname;
use File::Basename;
use Cwd;
use Carp;
use File::Spec; # for rel2abs
use Sys::Hostname;

use strict;

############## Wrappers around FileHandle methods #######################

sub openToRead {
    my ($f) = @_;
    if (! $f) {
	#die("Empty filename to read from\n");
	return *STDIN; # no arg means standard input
    } elsif ($f =~ /\.gz$/ || ($f =~ /\.gz\./ && $f !~ /\.gz\.touch$/)) {
	# allow .gz.sv1 etc ...
	return new FileHandle("gunzip -c $f |");
    } elsif (-s "$f.gz" && ! -s $f) {
	openToRead("$f.gz");
    } else {
	openSafely($f,"r","read from");
    }
}

sub openFromPipe {
    my ($exp,@files) = @_;
    $exp =~ s/\s+$//;
    $exp .= "|" unless (substr($exp,-1) eq "|");
    return openSafely($exp,undef,"pipe from");
}

sub openToPipe {
    my ($exp,@files) = @_;
    $exp =~ s/^\s+//;
    $exp =~ s/^\|//; # delete pipe at front
    return new FileHandle("| $exp");
}

sub openToWrite {
    my ($g) = @_;
    return *STDOUT unless ($g);
    my $dir = dirname($g);
    mkdir_p($dir);
    if ($g =~ /gz$/) {
	return new FileHandle("| gzip -c > $g");
    } else {
	return openSafely($g,"w","write to");
    }
}

# We convert to localEquivalent, but we don't do any lsrcp as
# we don't know when we would need to.
sub openToAppend {
    my ($f) = @_;
    openSafely($f,"a","append to");
}

sub openSafely {
    my ($f,$mode,$verb) = @_;
    if ($f) {
	if ($mode) {
	    return new FileHandle($f,$mode) || confess("Cannot $verb $f");
	} else {
	    return new FileHandle($f) || confess("Cannot $verb $f");
	}
    } else {
	return *STDOUT;
    }
}

sub openFromPaste {
    my (@files) = @_;
    foreach my $f (@files) {
	croak("openFromPaste: cannot find $f\n") unless (-f $f);
    }
    return openFromPipe("paste @files");
}

########### Read/write files between hashes, lists and aliases #############

sub fileToHash {
    my ($file,$pHash,$keyField,$valField) = @_;
    $keyField = 0 if ($keyField eq "");
    $valField = 1 if ($valField eq "");
    my $maxField = $valField;
    $maxField = $keyField if ($keyField > $maxField);
    my $fh = openToRead($file);
    while (<$fh>) {
	chomp;
	if (/^[^#]/) {
	    my @arr = split;
	    if ($#arr >= $maxField) {
		$pHash->{$arr[$keyField]} = $arr[$valField];
	    }
	}
    }
}

sub hashToFile {
    my ($pHash,$file,$mode) = @_;
    my $gh = openSafely($file,$mode,"write to");
    foreach my $key (sort keys %$pHash) {
	printf($gh "%s %s\n",$key,$pHash->{$key});
    }
    $gh->close();
}

sub fileToList {
    my ($file,$pList,$keyField,$valField) = @_;
    $keyField = 0 if ($keyField eq "");
    $valField = 1 if ($valField eq "");
    my $maxField = $valField;
    $maxField = $keyField if ($keyField > $maxField);
    my $fh = openToRead($file);
    while (<$fh>) {
	chomp;
	if (/^[^#]/) {
	    my @arr = split;
	    if ($#arr >= $maxField && $arr[$keyField] =~ /^\d+$/) {
		$pList->[$arr[$keyField]] = $arr[$valField];
	    }
	}
    }
}
    
sub listToFile {
    my ($pList,$file,$mode) = @_;
    my $gh = openSafely($file,$mode,"write to");
    foreach my $key (0 .. $#$pList) {
	if ($pList->[$key] ne "") {
	    printf($gh "%d %s\n",$key,$pList->[$key]);
	}
    }
    $gh->close();
}

sub readAliases {
    my ($f) = @_;
    return undef unless (-s $f);
    my %alias;
    my $fh = openToRead($f);
    my ($multi,$var);
    while (<$fh>) {
	chomp;
	unless (/^\s*\#/) {
	    next unless (/\S/);
	    if (/^\s/) {
		s/^\s+//;
		my @vals = split;
		push @{$alias{$var}},@vals;
		$multi = 1;
	    } else {
		my @vals;
		($var,@vals) = split;
		if (defined $var && @vals) {
		    my %h;
		    my $isHash = 1;
		    foreach my $val (@vals) {
			if ($val =~ /^(.*)=(.*)$/) {
			    $h{$1} = $2;
			} else {
			    $isHash = 0;
			    last;
			}
		    }
		    if ($isHash) {
			$alias{$var} = \%h;
		    } else {
			$alias{$var} = \@vals;
		    }
		    $multi = 1 if ($isHash || @vals > 1);
		}
	    }
	}
    }
    unless ($multi) {
	foreach my $var (keys %alias) {
	    $alias{$var} = $alias{$var}[0];
	}
    }
    return \%alias;
}

sub alias {
    my ($var,$pAli) = @_;
    return defined $pAli->{$var} ? $pAli->{$var} : $var;
}

#### Touchfile code ####

sub touch {
    my ($f) = @_;
    my $fh = new FileHandle($f,"w");
    if ($fh) {
	$fh->close();
    } else {
	croak("Can't touch $f\n");
    }
}

sub await {
    my ($file,$secs,$repDir) = @_;
    my $repFile = $file;
    $repFile = "$repDir/$file" if ($repDir);
    $secs = 60 unless ($secs > 0);
    my $nWaits = 0;
    while (-f "$file.touch" || ! -f $file) {
	if ($nWaits >= 10) {
	    my $gh = openToAppend($ENV{'HOME'} . "/await.log");
	    printf($gh "Awaiting %s at %s in command %s\n",
		   $repFile,scalar(localtime),
		   join(" ",$0,@ARGV));
	    $nWaits = 0;
	}
	print "Awaiting $repFile ...\n";
	sleep($secs);
	$nWaits ++;
    }
}


sub allFilesReady {
    my (@files) = @_;
    foreach my $f (@files) {
	return 0 if (-f "$f.touch" || ! -s $f);
    }
    return 1;
}

sub stamp {
    my ($f,$app) = @_;
    my $t = "$f.touch";
    # Fail if it's already there:
    return undef if (-f $t || (-f $f && !$app));
    # id is my hostname and process ID; should be unique...
    my $id = sprintf("%s.%d",hostname(),$$);
    my $gh = openToWrite($t);
    print $gh "$id\n";
    $gh->close();
    # Wait for any competitors
    sleep(3);
    return undef unless (-s $t);
    my $fh = openToRead($t);
    chomp ($_ = <$fh>);
    # Only succeed if we didn't get overwritten
    return ($id eq $_);
}

sub unstamp {
    my ($f,$nullOK) = @_;
    my $t = "$f.touch";
    croak("missing $t") unless ($nullOK || -f $t);
    unlink($t);
    return 1;
}

sub isComplete {
    my ($f,$rep) = @_;
    if (! -f $f) {
	report("$f: missing",1) if ($rep);
	return 0;
    }
    if (-f "$f.touch") {
	report("$f.touch: still exists",1) if ($rep);
	return 0;
    }
    return 1;
}

############ Estimate finish time for iterative process #############

sub estimatedFinishTime {
    # Started at time() = $t0; done $k out of $n iterations
    my ($t0,$k,$n) = @_;
    my $t1 = time();
    my $t2 = $t1; # wild guess for now
    if ($k > 0) {
	$t2 = $t1 + ($t1-$t0)*($n-$k)/$k;
    }
    return scalar(localtime($t2));
}

############# Progress reporter #####################

my $tReport; # static variable ...

sub report {
    my ($msg,$nl,$gh) = @_;
    $gh = *STDOUT unless ($gh);
    if ($msg) {
	$tReport = time();
	printf($gh "%s: %s",scalar(localtime($tReport)),replaceSgroot($msg));
	if ($nl) {
	    print $gh "\n";
	} else {
	    print $gh " ... ";
	}
    } else {
	my $d = time()-$tReport;
	my ($dStr,$units);
	if ($d >= 3600) {
	    $dStr = sprintf("%d:%2d:%2d",int($d/3600),int(($d-3600*int($d/3600))/60),$d-60*int($d/60));
	    $units = "hours";
	} elsif ($d >= 60) {
	    $dStr = sprintf("%d:%2d",int($d/60),$d-60*int($d/60));
	    $units = "minutes";
	} elsif ($d > 0) {
	    $dStr = $d;
	    $units = "seconds";
	} else {
	    $dStr = "no";
	    $units = "time";
	}
	$dStr =~ s/ /0/g;
	print $gh "done in $dStr $units\n";
    }
}

sub replaceSgroot {
    my ($msg) = @_;
    if ($msg =~ /^(.*) \/\S+\/sgroot(.*)$/) {
        return replaceSgroot("$1 \$SGR$2");
    } elsif ($msg =~ /\/\S+\/sgroot(.*)$/) {
        return replaceSgroot("\$SGR$1");
    } else {
	return $msg;
    }
}

############# min, max and list-comparison functions ###############
    
sub min {
    my ($x,$y) = @_;
    if (defined $y && $x > $y) {
	return $y;
    } elsif (defined $x) {
	return $x;
    } else {
	return $y;
    }
}
   
sub max {
    my ($x,$y) = @_;
    if (defined $y && $y > $x) {
	return $y;
    } elsif (defined $x) {
	return $x;
    } else {

	return $y;
    }
}

sub sameList {
    my ($pL1,$pL2) = @_;
    return 0 unless ($#$pL1 == $#$pL2);
    foreach my $k (0 .. $#$pL1) {
	return 0 unless ($pL1->[$k] eq $pL2->[$k]);
    }
    return 1;
}

############# File and directory manipulation #############

sub moveAside {
    my ($f) = @_;
    return unless (-f $f || -d $f);
    return if ($f =~ /\.sv\d+$/); # already an .sv file
    my $n=1;
    while (-s "$f.sv$n" || -d "$f.sv$n") {
	$n++;
    }
    rename($f,"$f.sv$n");
    return "$f.sv$n";
}    

sub newer {
    my ($f1,$f2) = @_;
    return 0 unless (-f $f1);
    return 1 unless (-f $f2);
    #croak("missing comparator in newer: $f2") unless (-f $f2);
    return 0 if ($f1 eq $f2); # equal is not newer!
    my @stat1 = stat($f1);
    my @stat2 = stat($f2);
    return ($stat1[9] > $stat2[9]);
}

sub mkdir_p {
    my ($dir) = @_;
    $dir = absoluteFileName($dir);
    unless (-d $dir) {
	my $dn = dirname($dir);
	mkdir_p($dn);
	mkdir($dir);
    }
}

sub absoluteFileName {
    my ($f) = @_;
    return File::Spec->rel2abs(File::Spec->canonpath($f));
}

sub absoluteReadLink {
    my ($f) = @_;
    if (-l $f) {
	return absoluteReadLink(sprintf("%s/%s",dirname($f),readlink($f)));
    } else {
	return $f;
    }
}

sub identicalFiles {
    my ($f,$g) = @_;
    return 0 unless (-f $f && -f $g);
    my $fh = openFromPipe("cmp $f $g 2>&1");
    my $msg = <$fh>;
    return !$msg;
}

############# All patterns of 0 and 1 of length n ###############

sub nonZeroBinaries {
    my ($n) = @_;
    my @ans = allBinaries($n);
    shift @ans; # to get rid of the 0000 one
    return @ans;
}

sub allBinaries {
    my ($n) = @_;
    if ($n == 0) {
	return ("");
    } else {
	my @tails = allBinaries($n-1);
	my @ans;
	foreach my $k (0,1) {
	    foreach my $tail (@tails) {
		push @ans,"$k$tail";
	    }
	}
	return @ans;
    }
}
		
###############

sub arch {
    my $fh = openFromPipe("uname -m"); # arch not defined on all machines
    $_ = <$fh>;
    chomp;
    return $_;
}

sub appendBeforeGZ {
    my ($f,$suf) = @_;
    if ($f =~ /\.gz$/) {
	$f =~ s/\.gz$/${suf}.gz/;
    } else {
	$f .= $suf;
    }
    return $f;
}

sub replaceBeforeGZ {
    my ($f,$from,$to) = @_;
    if ($f =~ /${from}\.gz$/) {
       $f =~ s/${from}\.gz$/${to}.gz/;
    } elsif ($f =~ /${from}$/) {
	$f =~ s/${from}$/$to/;
    } else {
	croak("Cannot find $from at end of $f");
    }
    return $f;
}

sub filesFromSmallest {
    my (@files) = @_;
    my %sz;
    foreach my $f (@files) {
	$sz{$f} = -s $f;
    }
    return sort {$sz{$a} <=> $sz{$b}} @files;
}

sub readFastaFile {
    my ($f) = @_;
    my $fh = openToRead($f);
    my (%seq,$buf,$chr);
    while (<$fh>) {
	chomp;
	if (/^>(\S+)/) {
	    $seq{$chr} = $buf if ($buf);
	    $chr = $1;
	    $buf = "";
	} else {
	    $buf .= $_;
	}
    }
    $seq{$chr} = $buf if ($buf);
    return \%seq;
}

sub writeFastaFile {
    my ($pSeq,$g) = @_;
    my $gh = openToWrite($g);
    foreach my $chr (sort keys %$pSeq) {
	my $seq = $pSeq->{$chr};
	print $gh ">$chr\n";
	my $len = length($seq);
	for (my $i=0; $i<$len; $i+=60) {
	    print $gh substr($seq,$i,60) . "\n";
	}
    }
    $gh->close();
}

sub avoidInitialDigit {
    my ($str) = @_;
    $str =~ s/^(\d+)/S\1/;
    my $lstr = $str;
    $lstr =~ tr/A-Z/a-z/;
    return ($str,$lstr);
}

sub printGFF {
    my ($gh,@arr) = @_;
    if (@arr > 9) {
	confess("Overlong GFF line: @arr");
    }
    unless (defined $gh) {
	confess("Undefined output handle for @arr");
    }
    foreach my $i (0 .. 8) {
	if (!defined $arr[$i]) {
	    $arr[$i] = '.';
	} elsif ($arr[$i] eq 'F') {
	    $arr[$i] = '+';
	} elsif ($arr[$i] eq 'C') {
	    $arr[$i] = '-';
	}
    }
    # Ensure 2nd field doesn't start with a digit
    #$arr[1] = avoidInitialDigit($arr[1]);
    print $gh join("\t",@arr) . "\n";
}

# Intended for Maq-style (multi-line sequence) Fastq files.

sub readQuad {
    my ($fh) = @_;
    my @a;
    # Set nucs header line
    my $line = <$fh>;
    return undef unless ($line);
    die("Expected line starting in @, got: $line") unless (substr($line,0,1) eq '@');
    chomp ($a[0] = $line);
    report("reading quad: $a[0]");
    # Read nucs
    while (1) {
	chomp (my $line = <$fh>);
	if (substr($line,0,1) eq '+') {
	    $a[2] = $line;
	    last;
	} else {
	    $a[1] .= $line;
	}
    }
    # Read same number of quals as we have nucs
    my $wanted = length($a[1]);
    while ($wanted > 0) {
	chomp (my $line = <$fh>);
	$wanted -= length($line);
	$a[3] .= $line;
    }
    report();
    return \@a;
}

sub END {
    my $tmpd = sprintf("/tmp/%s.%s",$ENV{USER},$ENV{LSB_JOBID});
    if (-d $tmpd) {
	sleep(15); # to give separate lsrcp processes etc time to finish
	`rm -rf $tmpd`;
    }
}

1;
